/******************* sysdep.c ********************
	contains the system dependent routines.
*************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifndef _POSIX_SOURCE
#define _POSIX_SOURCE
#endif
#include <sys/stat.h>

#include "types.h"
#include "files.h"
#include "sysdep.h"
#include "header.h"
#include "key.h"
#include "menu.h"

#define max(x,y) ((x)>(y)?(x):(y))

extern int topslot,hoffset,voffset;
int stopped;
char path[128];

void open_fwindow (int slot, int flag)
/********** open_fwindow
	open a window for the file.
	if flag==0, then use default window elements.
**********/
{	
	filetyp *ft=files+slot;
	int whandle,elements=0;

	/* Create and open a window for that file */
	strcpy(ft->info,"  Viewing");
	if (!flag)
	{	
		ft->flags.full=(params.full!=0);
		ft->elements=elements;
		ft->flags.indent=params.indent;
		ft->flags.blanks=params.blanks;
	}
	else elements=ft->elements;
	whandle=slot;
	ft->xold=xdesk; ft->yold=ydesk; ft->hold=hdesk; ft->wold=wdesk;
	ft->whandle=whandle;
	wsize(slot);
	topslot=slot;
}

void make_dup (char *path)
/********* make_dup
	make a duplicate of the file named in path.
**********/
{	
	char dupname[256];
	strcpy(dupname,path);
	strcat(dupname,".bak");
	remove(dupname);
	rename(path,dupname);
}
	
void save_file (int whandle)
/************* save_file
	save file in window.
	if whandle==0 then save top window.
	do not close the top window!
*************/
{	
	int slot;

	if (whandle==0) slot=topslot;
	else slot=find_slot(whandle);
	if (slot<0) return;

	if (!save_it(slot)) stopped=1;
	else files[slot].flags.updated=0;
	refresh_top();
}

void save_all (void)
{	int slot;
	for (slot=0; slot<MAXSLOTS; slot++)
		if (files[slot].flags.used && files[slot].flags.updated)
			if (save_it(slot)) files[slot].flags.updated=0;
	refresh_top();	
}

void new_top (void)
/********** new_top
	search for a new top window
***********/
{
	int slot;
	if (files[topslot].flags.used) return;
	for (slot=0; slot<MAXSLOTS; slot++)
		if (files[slot].flags.used)
		{
			topslot=slot; refresh_top(); return;
		}
	topslot=-1;
	refresh_top();
}

int close_file (int whandle)
/*********** close_file
	close a file. if updated, confirm with user.
	if whandle==0, take top window.
	return, if the file has been actually closed.
***********/
{	
	int slot,r;
	char *p,s[256];

	/* is it one of ours */
	stopped=0;
	if (whandle<0) return 0;
	slot=find_slot(whandle);
	if (slot<0) return 0;
	
	/* close it */
	if (files[slot].flags.updated)
	{	
		p=files[slot].name;
		r=max(0,(int)strlen(p)-20);
		p+=r;
		sprintf(s,"[2][Save|...%s][yes|no|cancel]",p);
		r=form_alert(1,s);
		if (r==1) { if (!save_it(slot)) stopped=1; }
		else if (r==3) stopped=1;
		if (stopped) return 0;
	}
	freset(slot);
	new_top();
	return 1;

}

void overload (void)
/*********** overload
	close top window and load another file for top window.
***********/
{	
	if (topslot>=0) close_file(topslot);
	load_file();
	new_top();
}

void save_dup (void)
/********** save_dup
	make a duplicate of the top window and save the top window.
**********/
{	
	int slot,whandle;

	whandle=topslot;
	
	/* is it one of ours */
	slot=find_slot(whandle);
	if (slot<0) return;
	
	make_dup(files[slot].name);
	save_file(whandle);
}

void save_as (void)
/********* save_as
	ask a name from user and save top window with that name.
	do not close the top window!
*********/
{	
	int slot;
	static char string[256]="";
	char oldname[256]="";

	slot=get_top();
	if (slot<0) return;
	
	/* select file */
	select_file("Save as : ",string);
	if (string[0]==0) return;

	strcpy(oldname,files[slot].name);
	strcpy(files[slot].name,string);
	
	save_file(files[slot].whandle);
	
	if(form_alert(1,"[1][Change internal|file name?][Yes|No]")==2)
	{	strcpy(files[slot].name,oldname);
		refresh_top();
	}
}

void del_file (void)
/*********
	delete a file on disk. confirm with user.
*********/
{	
}

void error (char *msg)
/********** error
	show an error message 
**********/
{	
	char string[256];
	sprintf(string,"[3][%s][Sorry]",msg);
	form_alert(1,string);
}

void error1 (char *msg, char *string)
/********** error1 
	show an error message. msg must contain a %s 
**********/
{	
	char s[256]="[3][";
	sprintf(s+strlen(s),msg,string);
	strcat(s,"][Sorry]");
	form_alert(1,string);
}

void select_file (char *prompt, char *string)
/******** select_file (system dependent!)
	file selection. keeps the selected path.
*********/
{	
	binput(prompt,string);
	while (string[0]==' ' || string[0]==9) strcpy(string,string+1);
	if (!string[0]) refresh_top();
}
	
void fredraw (int slot)
/*********** fredraw
	redraw a slot.
***********/
{	
	if (topslot==slot) refresh_top();
}

int get_top (void)
/********* get_top
	get top window slot.
*********/
{
	return topslot;
}

void wind_top (int slot)
/********* wind_top
	make the slot top window.
*********/
{	
	if (files[slot].flags.used)
	{	topslot=slot;
		refresh_top();
	}
}

int memsave (char *path, char *start, size_t size)
/*********
	save a portion of memory to the file named in path.
	return 0 on error.
*********/
{	
	FILE *out;
	out=fopen(path,"wb");
	if (!out)
	{	error("Could not open that file!");
		return 0;
	}
	fwrite(start,(size_t)1,size,out);
	if (ferror(out))
	{
		error("Save error!");
		return 0;
	}
	fclose(out);
	return 1;
}

int memload (char *path, char *start, size_t size)
/***********
	load a file from path.
	return 0 on error
***********/
{	
	FILE *in;
	in=fopen(path,"rb");
	if (!in)
	{	error("Could not open that file!");
		return 0;
	}
	fread(start,(size_t)1,size,in);
	if (ferror(in))
	{
		error("Load error!");
		return 0;
	}
	fclose(in);
	return 1;
}

int search_file (char *path, char name[], size_t *size, int *mode)
/*********** search_file
	search a file named in path. if there is any name[] should
	contain the file name, size the file size.
	return 1 if there is a file.
***********/
{
	struct stat buf;
	strcpy(name,path);
	if (stat(path,&buf)) return 0;
	*size=buf.st_size;
	*mode=buf.st_mode;
	return 1;
}

void set_mode (char *path, int mode)
{	if (mode) chmod(path,mode);
}

/****************** program end *********************/

int end_program (void)
/******** end_program
	close and save all files, if updated.
	return, if all files could be closed.
********/
{
	int slot;
	stopped=0;
	if (params.config) save_config();
	for (slot=0; slot<MAXSLOTS; slot++)
	{	if (files[slot].flags.used) 
			if (!close_file(files[slot].whandle)) return 0;
		if (stopped) { new_top(); return 0; }
	}
	return 1;
}

/********************** load and save configuration **********/

void save_config (void)
/********* save_config
	save the configuration.
*********/
{	
	FILE *out;
	filetyp *ft;
	int i;
	out=fopen("editor.cfg","w");
	if (!out)  { error("Couldn't open editor.cfg!"); return; }
	fprintf(out,"reneedit91\n");
	fprintf(out,"%d %d %d %d %d %d %d %d %d\n%ld %ld %ld\n",
		params.full,params.tabsize,params.linelength,
		params.indent,params.blanks,params.dup,params.arguments,
		params.config,params.paral,
		params.reserve,params.freserve,params.memfree);
	for (i=0; i<MAXSLOTS; i++)
	{	ft=files+i;
		if (ft->flags.used)
			fprintf(out,"1 %s\n%d %d %d %d %d %d %d\n"
				"%d %d %d %d %d %d %d\n"
				"%d %d %d\n%d %ld\n",
				ft->name,
				ft->flags.info, ft->flags.readonly,
				ft->flags.full,ft->flags.indent,ft->flags.insert,
				ft->flags.blanks,ft->flags.wordwrap,
				ft->x,ft->y,ft->w,ft->h,ft->xl,ft->wl,ft->hl,
				ft->left,ft->tabs,ft->linelength,
				ft->elements,(long)(ft->cursor.line-ft->start));
	}
	fprintf(out,"%d",0);
	fclose(out);
}

void load_config (void)
/********* load_config
	load the configuration. assumes, that all slots are unused.
*********/
{	
	FILE *in;
	filetyp *ft;
	int i,flag,j;
	char s[128];
	long position;
	in=fopen("editor.cfg","r");
	if (!in)  return;
	fscanf(in,"%s",s);
	if (strcmp(s,"reneedit91"))  
	{ error("Illegal format|in config file!"); return; }
	fscanf(in,"%d%d%d%d%d%d%d%d%d%ld%ld%ld",
		&params.full,&params.tabsize,&params.linelength,
		&params.indent,&params.blanks,&params.dup,&params.arguments,
		&params.config,&params.paral,
		&params.reserve,&params.freserve,&params.memfree);
	for (i=0; i<MAXSLOTS; i++)
	{	ft=files+i;
		if (fscanf(in,"%d",&flag)!=1) break;
		if (!flag) break;
		fscanf(in,"%s",s); /* name of file */
		fscanf(in,"%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%d%ld",
			&ft->flags.info, &ft->flags.readonly,
			&ft->flags.full,&ft->flags.indent,&ft->flags.insert,
			&ft->flags.blanks,&ft->flags.wordwrap,
			&ft->x,&ft->y,&ft->w,&ft->h,&ft->xl,&ft->wl,&ft->hl,
			&ft->left,&ft->tabs,&ft->linelength,
			&ft->elements,&position);
		load_it(s,i,1);
		ft->topl=ft->cursor.adr=ft->cursor.line=
			prev_line(ft->start+position,ft);
		ft->cursor.x=0;
		for (j=0; j<5; j++)
			ft->topl=prev_line(ft->topl,ft);
	}
	fclose(in);
}

#define upd(x) x=p+((x)-ramstart)

void extend_memory (void)
{	char input[256];
	static char extend[256]="128";
	char *p;
	long size;
	filetyp *ft;
	int i;
	sprintf(input,"%ld KB (%ld KB free) plus",
		(long)(ramend-ramstart)/1024,(long)(ramend-ramfree)/1024);
	binput(input,extend);
	if (!extend[0] || extend[0]=='0') goto nothing;
	sscanf(extend,"%ld",&size);
	if (size<=0) goto nothing;
	size=1024*size+(ramend-ramstart);
	p=realloc(ramstart,size);
	if (!p)
	{	form_alert(1,"[1][Not enough memory|available][Sorry]");
		goto nothing;
	}
	if (p!=ramstart)
	{	for (i=0; i<MAXSLOTS; i++)
		{	ft=files+i;	
			upd(ft->start);
			upd(ft->end);
			upd(ft->fend);
			upd(ft->topl);
			upd(ft->cursor.line);
			upd(ft->cursor.adr);
			if (ft->gap.open)
			{	upd(ft->gap.start);
				upd(ft->gap.end);
			}
		}
		/* change blocks */
		if (block.slot>=0)
		{	upd(block.start);
			upd(block.end);
		}
		/* change markers */
		for (i=0; i<MAXMARKER; i++)
			if ((marker[i].slot>=0))
				upd(marker[i].line);
		/* correct undo buffer */
		undo.slot=-1;
		ramstart=p;
	}
	ramend=ramstart+size;
	nothing: refresh_top();
}
